#!/usr/bin/env python
# -*- coding:utf-8 -*-
from inspect import getfullargspec
from typing import TypeVar

from ..exceptions import CallException

T = TypeVar("T")


def run_call_back(call_func: callable, origin_func: callable, args: tuple, kwargs: dict, origin_result: T = None) -> T:
    try:
        if callable(call_func):
            new_kwargs = get_func_params(origin_func, args, kwargs)
            spec = getfullargspec(call_func)
            params = {}
            if "func_params" in spec.args or "func_params" in spec.kwonlyargs:
                params["func_params"] = new_kwargs
            if "func_return" in spec.args or "func_return" in spec.kwonlyargs:
                params["func_return"] = origin_result
            return call_func(**params)
    except Exception as e:
        raise CallException(f"call back exception: {str(e)}")


def get_func_params(func: callable, args: tuple, kwargs: dict) -> dict:
    new_params = {}  # 将所有参数以dict形式保存
    arg_spec = getfullargspec(func)

    tmp_arg_names = arg_spec.args
    tmp_arg_values = __copy_args(args)
    tmp_kwarg_kvs = __copy_kwargs(kwargs)
    if len(tmp_arg_names) > 0 and len(tmp_arg_values) > 0:
        first_value = tmp_arg_values[0]
        if func.__qualname__.split(".")[0] == first_value.__class__.__name__:  # 即为类方法或实例方法,添加到字典中
            new_params[tmp_arg_names.pop(0)] = tmp_arg_values.pop(0)
    choice_args_values = tmp_arg_values[:len(tmp_arg_names)]
    no_choice_values = tmp_arg_values[-len(tmp_arg_names):]

    new_params.update(dict(zip(tmp_arg_names, choice_args_values)))  # 添加位置参数kv

    kw_defaults = arg_spec.kwonlydefaults
    if not kw_defaults:
        kw_defaults = {}
    kw_defaults_keys = kw_defaults.keys()
    must_need_value_keys = [i for i in arg_spec.kwonlyargs if i not in kw_defaults_keys]
    for key in kw_defaults_keys:  # 添加kwargs，如果用户传了，使用用户的参数，否则用默认值
        if key in kwargs:
            new_params[key] = kwargs[key]
            del tmp_kwarg_kvs[key]
        else:
            new_params[key] = kw_defaults[key]
    for key in must_need_value_keys:  # 这类关键字参数没有默认值，从剩余的参数中依次赋值
        if key not in new_params and key in kwargs:
            new_params[key] = kwargs[key]
        else:
            new_params[key] = None
    if arg_spec.varargs:
        new_params[arg_spec.varargs] = no_choice_values  # 将剩余的args参数复制给函数形参的args
    if arg_spec.varkw:
        new_params[arg_spec.varkw] = tmp_kwarg_kvs  # 将剩余的kwargs参数赋值给函数的形参kwargs
    return new_params


def __copy_args(args: tuple or list) -> list:
    tmp_args = []
    tmp_args_append = tmp_args.append
    for arg in args:
        tmp_args_append(arg)
    return tmp_args


def __copy_kwargs(kwargs: dict) -> dict:
    tmp_kwargs = {}
    for k, v in kwargs.items():
        tmp_kwargs[k] = v
    return tmp_kwargs
