#!/usr/bin/env python
# -*- coding:utf-8 -*-
from functools import reduce
from itertools import groupby, chain
from random import choice
from typing import TypeVar, Iterable, Callable, List, Dict, Any, Optional

T = TypeVar('T')
R = TypeVar("R")
K = TypeVar('K')
V = TypeVar('V')


class _AbstractList(List[T]):

    def __init__(self, *args: T, iterable: Iterable[T] = None):
        if args:
            super().__init__(args)
        elif iterable:
            super().__init__(iterable)
        else:
            super().__init__([])

    def map(self, func: Callable[[T], R]) -> '_AbstractList[R]':
        """
        A stream of element type T is converted to a stream of element type R through a Function.
        """
        return _AbstractList(iterable=(func(element) for element in self))

    def flat_map(self, func: Callable[[T], '_AbstractList[R]'] = lambda x: x) -> '_AbstractList[R]':
        """
        Flattening the flow, using a Function to convert each element in a stream with an element type T into
        a stream with an element type R, and then merging the converted streams.
        """
        return _AbstractList(iterable=chain.from_iterable(map(func, self)))

    def filter(self, func: Callable[[T], bool]) -> '_AbstractList[T]':
        """
        Filter the elements in the stream, retaining only those elements that meet the criteria specified by Predicate:
        """
        return _AbstractList(iterable=(element for element in self if func(element)))

    def distinct(self) -> '_AbstractList[T]':
        """
        Filter to remove duplicate elements by hashCode() and equals of the elements generated by the stream
        """
        return _AbstractList(iterable=set(self))

    def limit(self, size: int) -> '_AbstractList[T]':
        """
        Truncates the stream so that it contains only a specified number of elements at most.
        """
        if len(self) <= size:
            return _AbstractList(iterable=self)
        else:
            return _AbstractList(iterable=self[:size])

    def skip(self, size: int) -> '_AbstractList[T]':
        """
        Skipping elements, returning a stream with the first n elements thrown away, and returning an empty stream
        if there are less than n elements in the stream, complementary to limit(n).
        """
        if len(self) <= size:
            return _AbstractList(iterable=[])
        else:
            return _AbstractList(iterable=self[size:])

    def sorted(self, key=None, reverse: bool = False) -> '_AbstractList[T]':
        """
        Sort the stream.
        """
        return _AbstractList(iterable=sorted(self, key=key, reverse=reverse))

    def take_while(self, predicate: Callable[[T], bool]) -> '_AbstractList[T]':
        """
        Elements that satisfy the specified predicate are retained from the beginning of
        the original stream until the first element that does not satisfy the Predicate is encountered.
        """
        tmp = []
        tmp_append = tmp.append
        for element in self[:]:
            if predicate(element):
                tmp_append(element)
            else:
                break
        return _AbstractList(iterable=tmp)

    def drop_while(self, predicate: Callable[[T], bool]) -> '_AbstractList[T]':
        """
        Removes elements that satisfy the specified predicate from the beginning of
        the original stream until the first element that does not satisfy Predicate is encountered.
        """
        tmp = self[:]
        for i in range(len(self)):
            if predicate(self[i]):
                tmp.remove(self[i])
            else:
                break
        return tmp

    def for_each(self, func: Callable[[T], None]) -> None:
        """
        Traversal, custom actions
        """
        for i in self:
            func(i)

    def all_match(self, func: Callable[[T], bool]) -> bool:
        """
        Check that all elements match
        """
        return all(map(func, self))

    def any_match(self, func: Callable[[T], bool]) -> bool:
        """
        Check that at least one element matches
        """
        return any(map(func, self))

    def none_match(self, func: Callable[[T], bool]) -> bool:
        """
        Check that not all elements are matched
        :param func:
        :return:
        """
        return not self.any_match(func)

    def find_first(self) -> Optional[T]:
        """
        Returns the first element
        """
        if len(self) > 0:
            return self[0]
        return None

    def find_any(self) -> Optional[T]:
        """
        Returns any element in the current stream
        """
        if len(self) > 0:
            return choice(self)
        return None

    def reduce(self, func: Callable[[T, T], T], initial: T = None) -> Optional[T]:
        if initial is not None:
            return reduce(func, self, initial)
        else:
            try:
                return reduce(func, self)
            except TypeError:
                return None

    def count(self, obj: Any = None) -> int:
        """
        Returns the total number of elements in the stream
        :param obj: Elements with statistics, if None counts all
        """
        if obj:
            return super().count(obj)
        return len(self)

    def sum(self) -> int or float:
        """
        Sum the elements of the list
        """
        return sum(self)

    def max(self, key: Callable[[T], Any] = lambda x: x, default: T = None) -> Optional[T]:
        """
        Returns the maximum value in the stream
        """
        return max(self, key=key, default=default)

    def min(self, key: Callable[[T], Any] = lambda x: x, default: T = None) -> Optional[T]:
        """
        Returns the smallest value in the stream
        """
        return min(self, key=key, default=default)

    def group(self, func: callable) -> Dict[K, '_AbstractList[T]']:
        """
        Group streams
        """
        tmp = {}
        for key, group in groupby(self, key=func):
            if key in tmp:
                tmp[key].extend(_AbstractList(iterable=group))
            else:
                tmp[key] = _AbstractList(iterable=group)
        return tmp

    def to_dict(self, k: Callable[[T], K], v: Callable[[T], V], overwrite: bool = True) -> Dict[K, V]:
        """
        Converts the stream into a dictionary
        :param k: Generate dictionary keys
        :param v: Generate the dictionary value
        :param overwrite: Whether to override when present
        """
        tmp = {}
        for i in self:
            key = k(i)
            value = v(i)
            if overwrite:
                tmp[key] = value
            else:
                if key not in tmp:
                    tmp[key] = value
        return tmp

    def get(self, index: int) -> Optional[T]:
        """
        Gets the element that specifies the subscript
        :param index:
        :return:
        """
        if 0 <= index <= len(self):
            return self[index]
        return None
