#!/usr/bin/env python
# -*- coding:utf-8 -*-
from datetime import datetime
from pathlib import Path
from threading import Lock

from ..converter.storage import StorageUnit
from ..log.level import LogLevel


class __LogConfig(object):
    """
    Log global configuration, must be loaded first
    """
    __lock__ = Lock()
    __instance = None

    def __new__(cls, *args, **kwargs):
        with cls.__lock__:
            if not cls.__instance:
                cls.__instance = object.__new__(cls)
        return cls.__instance

    def __init__(self):
        self.__dir: Path = Path.cwd().joinpath("logs").absolute()
        self.__level: LogLevel = LogLevel.NOSET
        self.__level_console = LogLevel.CRITICAL
        self.__level_file = LogLevel.CRITICAL
        self.__format: str = "[%(asctime)s]-[%(process)s]-[%(thread)s]-[%(filename)s %(lineno)s]-[%(levelname)s] %(message)s"
        self.__coding: str = "utf-8"
        self.__name: str = f"simplebox-{datetime.now().strftime('%Y-%m-%d')}.log"
        self.__max_bytes: int = StorageUnit.MB.of(100).to(StorageUnit.BYTE).integer()
        self.__backup_count: int = 10
        self.__path: str = str(self.__dir.joinpath(self.__name).absolute())
        self.__banner: bool = True
        self.__off_file: bool = False
        self.__off_console: bool = False
        self.__off: bool = False
        if not self.__dir.exists():
            self.dir.mkdir(parents=True)

    @property
    def dir(self) -> Path:
        return self.__dir

    @dir.setter
    def dir(self, value: str):
        path = Path(value)
        if not path.exists():
            path.mkdir(parents=True)
        self.__dir = value
        self.__path = Path(self.__dir).joinpath("logs", self.__name)

    @property
    def level(self) -> LogLevel:
        return self.__level

    @level.setter
    def level(self, value: LogLevel):
        self.__level = value

    @property
    def level_console(self) -> LogLevel:
        return self.__level_console

    @level_console.setter
    def level_console(self, value: str):
        if isinstance(value, str):
            self.__level_console = value

    @property
    def level_file(self) -> LogLevel:
        return self.__level_file

    @level_file.setter
    def level_file(self, value: str):
        if isinstance(value, str):
            self.__level_file = value

    @property
    def format(self) -> str:
        return self.__format

    @format.setter
    def format(self, value: str):
        self.__format = value

    @property
    def coding(self) -> str:
        return self.__coding

    @coding.setter
    def coding(self, value: str):
        self.__coding = value

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str):
        self.__name = value
        self.__path = Path(self.dir).joinpath(self.__name)

    @property
    def max_bytes(self) -> int:
        return self.__max_bytes

    @max_bytes.setter
    def max_bytes(self, value: int):
        self.__max_bytes = value

    @property
    def backup_count(self) -> int:
        return self.__backup_count

    @backup_count.setter
    def backup_count(self, value: int):
        self.__backup_count = value

    @property
    def path(self) -> str:
        return self.__path

    @property
    def banner(self) -> bool:
        return self.__banner

    @banner.setter
    def banner(self, value: bool):
        if isinstance(value, bool):
            self.__banner = value

    @property
    def off_file(self) -> bool:
        return self.__off_file

    @off_file.setter
    def off_file(self, value: bool):
        if isinstance(value, bool):
            self.__off_file = value

    @property
    def off_console(self) -> bool:
        return self.__off_console

    @off_console.setter
    def off_console(self, value: bool):
        if isinstance(value, bool):
            self.__off_console = value

    @property
    def off(self) -> bool:
        return self.__off

    @off.setter
    def off(self, value: bool):
        if isinstance(value, bool):
            self.__off = value
