#!/usr/bin/env python
# -*- coding:utf-8 -*-
import json
import re
import traceback
from copy import deepcopy
from enum import Enum
from functools import wraps
from inspect import getfullargspec
from pathlib import Path
from typing import List, Dict, TypeVar
from urllib.parse import urljoin

import requests
from requests import Response
from ..config import RestConfig
from ..utils.enums import get_by_value

from ..utils.tools import check_none, get_enum
from ..exceptions import HttpException
from ..log.factory import LoggerFactory

T = TypeVar("T")

optional_args_keys = ["params", "data", "json", "headers", "cookies", "files", "auth", "timeout", "allow_redirects",
                      "proxies", "verify", "collection", "cert"]

pattern = re.compile(f"^http|https?:/{2}\\w.+$")

logger = LoggerFactory.get_logger()


class Method(Enum):
    GET = "GET"
    OPTIONS = "OPTIONS"
    HEAD = "HEAD"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


class Rest(object):
    def __init__(self, file: str = None, server_name: str = None, host: str = None, only_body: bool = True,
                 check_status: bool = False, encoding: str = "utf-8"):
        """
        Build a request client.
        :param file: The path where the interface configuration file is stored.
                     configuration format：
                        [
                          {
                            "serverName": "s1",
                            "serverHost": "http://localhost1",
                            "desc": "",
                            "apis": [
                              {
                                "apiName": "user",
                                "apiPath": "/user",
                                "httpMethod": "post",
                                "headers": {"Content-type": "multipart/form-data"},
                                "desc": ""
                              }
                            ]
                          },
                          {
                            "serverName": "s2",
                            "serverHost": "http://localhost2",
                            "desc": "",
                            "apis": [
                              {
                                "apiName": "admin",
                                "apiPath": "/admin",
                                "httpMethod": "get",
                                "desc": ""
                              }
                            ]
                          }
                        ]
        :param server_name: Service name, which allows you to read interface information from the interface configuration file.
        """
        self.__only_body: bool = only_body
        self.__check_status: bool = check_status
        self.__encoding: str = encoding
        self.__server_name: str = server_name
        self.__server_list: List[Dict[str, str]] = []
        self.__server: Dict[str, str] = {}
        self.__host: str = host
        if file:
            path = Path(file)
            if not path.is_absolute():
                path = Path.cwd().joinpath(file)
            if not path.exists():
                raise RuntimeError(f"not found file: {path}")
            with open(path.absolute(), "r") as f:
                self.__server_list = json.load(f)

    def request(self, api_name: str = None, server_name: str = None, host: str = None, api: str = None,
                method: Method or str = None, allow_redirection: bool = RestConfig.allow_redirection,
                headers: dict = None,
                check_status: bool = RestConfig.check_status,
                encoding: str = RestConfig.encoding, only_body: bool = RestConfig.only_body):
        """
        Configure the interface information
        :param only_body: only need response body, if you need origin response, set false, default true
        :param encoding: parse response's text or content encode
        :param check_status: check http response status, default false
        :param api_name: Specify the API name, if empty while use function name as api name
        :param server_name: service name, which overrides the server_name of the instance.
                            If it is blank and the instance server_name is also blank,
                            the class name is used as the server name
        :param host: interface domain name, which is used first
        :param api: service http interface, which takes precedence over this parameter when specified
        :param method: interface request method, which is used in preference after specified
        :param allow_redirection: Whether to automatically redirect, the default is
        :param headers: custom http request header, if allow_redirection parameter is included,
        the allow_redirection in the header takes precedence
        """

        def __inner(func):
            @wraps(func)
            def __wrapper(*args, **kwargs):
                self.__request(func=func, kwargs=kwargs, api_name=api_name, server_name=server_name, host=host, api=api,
                               method=method, allow_redirection=allow_redirection, headers=headers,
                               check_status=check_status, encoding=encoding, only_body=only_body)
                return func(*args, **kwargs)

            return __wrapper

        return __inner

    def get(self, api_name: str = None, server_name: str = None, host: str = None, api: str = None,
            allow_redirection: bool = RestConfig.allow_redirection, headers: dict = None,
            check_status: bool = RestConfig.check_status,
            encoding: str = RestConfig.encoding, only_body: bool = RestConfig.only_body):
        def __inner(func):
            @wraps(func)
            def __wrapper(*args, **kwargs):
                self.__request(func=func, kwargs=kwargs, api_name=api_name, server_name=server_name, host=host, api=api,
                               method=Method.GET, allow_redirection=allow_redirection, headers=headers,
                               check_status=check_status, encoding=encoding, only_body=only_body)
                return func(*args, **kwargs)

            return __wrapper

        return __inner

    def post(self, api_name: str = None, server_name: str = None, host: str = None, api: str = None,
             allow_redirection: bool = RestConfig.allow_redirection, headers: dict = None,
             check_status: bool = RestConfig.check_status,
             encoding: str = RestConfig.encoding, only_body: bool = RestConfig.only_body):
        def __inner(func):
            @wraps(func)
            def __wrapper(*args, **kwargs):
                self.__request(func=func, kwargs=kwargs, api_name=api_name, server_name=server_name, host=host, api=api,
                               method=Method.POST, allow_redirection=allow_redirection, headers=headers,
                               check_status=check_status, encoding=encoding, only_body=only_body)
                return func(*args, **kwargs)

            return __wrapper

        return __inner

    def put(self, api_name: str = None, server_name: str = None, host: str = None, api: str = None,
            allow_redirection: bool = RestConfig.allow_redirection, headers: dict = None,
            check_status: bool = RestConfig.check_status,
            encoding: str = RestConfig.encoding, only_body: bool = RestConfig.only_body):
        def __inner(func):
            @wraps(func)
            def __wrapper(*args, **kwargs):
                self.__request(func=func, kwargs=kwargs, api_name=api_name, server_name=server_name, host=host, api=api,
                               method=Method.PUT, allow_redirection=allow_redirection, headers=headers,
                               check_status=check_status, encoding=encoding, only_body=only_body)
                return func(*args, **kwargs)

            return __wrapper

        return __inner

    def delete(self, api_name: str = None, server_name: str = None, host: str = None, api: str = None,
               allow_redirection: bool = RestConfig.allow_redirection, headers: dict = None,
               check_status: bool = RestConfig.check_status,
               encoding: str = RestConfig.encoding, only_body: bool = RestConfig.only_body):
        def __inner(func):
            @wraps(func)
            def __wrapper(*args, **kwargs):
                self.__request(func=func, kwargs=kwargs, api_name=api_name, server_name=server_name, host=host, api=api,
                               method=Method.DELETE, allow_redirection=allow_redirection, headers=headers,
                               check_status=check_status, encoding=encoding, only_body=only_body)
                return func(*args, **kwargs)

            return __wrapper

        return __inner

    def patch(self, api_name: str = None, server_name: str = None, host: str = None, api: str = None,
              allow_redirection: bool = RestConfig.allow_redirection, headers: dict = None,
              check_status: bool = RestConfig.check_status,
              encoding: str = RestConfig.encoding, only_body: bool = RestConfig.only_body):
        def __inner(func):
            @wraps(func)
            def __wrapper(*args, **kwargs):
                self.__request(func=func, kwargs=kwargs, api_name=api_name, server_name=server_name, host=host, api=api,
                               method=Method.PATCH, allow_redirection=allow_redirection, headers=headers,
                               check_status=check_status, encoding=encoding, only_body=only_body)
                return func(*args, **kwargs)

            return __wrapper

        return __inner

    def head(self, api_name: str = None, server_name: str = None, host: str = None, api: str = None,
             allow_redirection: bool = RestConfig.allow_redirection, headers: dict = None,
             check_status: bool = RestConfig.check_status,
             encoding: str = RestConfig.encoding, only_body: bool = RestConfig.only_body):
        def __inner(func):
            @wraps(func)
            def __wrapper(*args, **kwargs):
                self.__request(func=func, kwargs=kwargs, api_name=api_name, server_name=server_name, host=host, api=api,
                               method=Method.HEAD, allow_redirection=allow_redirection, headers=headers,
                               check_status=check_status, encoding=encoding, only_body=only_body)
                return func(*args, **kwargs)

            return __wrapper

        return __inner

    def options(self, api_name: str = None, server_name: str = None, host: str = None, api: str = None,
                allow_redirection: bool = RestConfig.allow_redirection, headers: dict = None,
                check_status: bool = RestConfig.check_status,
                encoding: str = RestConfig.encoding, only_body: bool = RestConfig.only_body):
        def __inner(func):
            @wraps(func)
            def __wrapper(*args, **kwargs):
                self.__request(func=func, kwargs=kwargs, api_name=api_name, server_name=server_name, host=host, api=api,
                               method=Method.OPTIONS, allow_redirection=allow_redirection, headers=headers,
                               check_status=check_status, encoding=encoding, only_body=only_body)
                return func(*args, **kwargs)

            return __wrapper

        return __inner

    def __request(self, func: callable, kwargs: dict, api_name: str = None, server_name: str = None, host: str = None,
                  api: str = None, method: Method or str = None, allow_redirection: bool = True, headers: dict = None,
                  check_status: bool = None, encoding: str = None, only_body: bool = True):
        func_origin_args = getfullargspec(func).args
        logger.info(f"{'Rest Start'.center(41, '*')}")
        if "response" not in func_origin_args:
            raise RuntimeError(f"function {func.__name__} need 'response' args, ex: {func.__name__}(response) "
                               f"or {func.__name__}(response=None)")
        headers_: dict = headers
        server_name_: str = self.__build_server_name(server_name, func)
        api_name_: str = self.__build_api_name(api_name, func)
        host_dict: dict = self.__get_server_dict(server_name_)
        host_: str = self.__get_host(host, host_dict)
        self.__check_domain(host_)
        api_: str = self.__get_api_info(api, host_dict, api_name_, "apiPath", "not found api")
        method_: str = self.__get_api_info(self.__get_request_method(method),
                                           host_dict, api_name, "httpMethod", "not found http method")
        headers_: dict = self.__get_api_info(headers_, host_dict, api_name, "headers", "not found http header")
        check_none(api_)
        check_none(get_enum(Method, method_.upper()))
        optional_args: dict = self.__build_optional_args(kwargs)
        self.__build_header(optional_args, allow_redirection, method_.upper(), headers_)
        url: str = urljoin(host_, api_)
        check_status_: bool = self.__check_status if not check_status else check_status
        encoding_: str = self.__encoding if not encoding else encoding
        only_body_: bool = self.__only_body if not only_body else only_body
        # noinspection PyBroadException
        try:
            resp: Response or None = self.__action(method_.lower(), url, **optional_args)
        except BaseException as e:
            e.__init__(f"An exception occurred during the http request process: url is {host_}{api_}")
            logger.error(f"An exception occurred when a request was sent without a response:\n"
                         f"{traceback.format_exc()}")
            raise

        # noinspection PyBroadException
        try:
            logger.info(f"http response => http code: {resp.status_code}, "
                        f"content: {resp.content.decode(encoding_)[:50]}...")
            if check_status_:
                if 200 > resp.status_code or resp.status_code >= 300:
                    logger.error(f"check http status code is not success: {resp.status_code}")
                    raise HttpException(f"http status code is not success: {resp.status_code}")
            if only_body_:
                kwargs["response"] = resp.json()
            else:
                kwargs["response"] = resp
        except BaseException:
            if resp:
                logger.error(
                    f"'{url}' response can't parse dict, response: {resp.content.decode(encoding_)[:50]}, "
                    f"exception: \n{traceback.format_exc()}")
                kwargs["response"] = resp.content.decode(encoding_)
            else:
                kwargs["response"] = None
        finally:
            logger.info(f"{'Rest End'.center(43, '*')}")

    def __get_server_dict(self, name: str) -> dict:
        if self.__server_list:
            for server in self.__server_list:
                if server.get("serverName") == name:
                    return server
        return {}

    def __get_host(self, host: str, host_dict: dict) -> str:
        host_: str = host
        if not host_:
            host_: str = self.__host
        if not host_:
            host_: str = host_dict.get("serverHost")
        return host_

    def __build_server_name(self, server_name: str, func: callable) -> str:
        if isinstance(server_name, str) and server_name.strip() != "":
            return server_name
        if isinstance(self.__server_name, str) and self.__server_name.strip() != "":
            return self.__server_name
        return func.__qualname__.split(".")[0]

    @staticmethod
    def __get_request_method(method: Method or str) -> str:
        if isinstance(method, Method):
            return method.value
        elif isinstance(method, str):
            return get_by_value(Method, method, Method.GET).value
        else:
            return Method.GET.value

    @staticmethod
    def __action(http_method: str, url: str, **kwargs) -> Response:
        logger.info(f"http request => url: {url}, method: {http_method}, arguments: {kwargs}")
        kwargs["url"] = url
        action = getattr(requests, http_method, None)
        if action:
            return action(**kwargs)
        else:
            raise HttpException(f"unknown http method '{http_method}'")

    @staticmethod
    def __get_api_info(default: T, host_dict: dict, api_name, key: str, err_msg: str) -> T:
        """
        Take the value from the API dictionary, and use origin when it can't
        """
        default_: dict = default
        if not default_ and "apis" in host_dict:
            api_list: List[Dict] = host_dict.get("apis")
            if not api_list:
                raise RuntimeError(err_msg)
            for api in api_list:
                if isinstance(api, dict) and api.get("apiName") == api_name:
                    return api.get(key)
        return default_

    @staticmethod
    def __check_domain(domain):
        if not pattern.match(domain):
            raise RuntimeError(f"invalid host: {domain}")

    @staticmethod
    def __build_header(all_params: dict, allow_redirection_: bool = True, method: str = Method.GET.value,
                       headers: dict = None):
        headers_: dict = all_params.get("headers")
        if not headers_:
            if method == Method.POST.value or method == Method.PUT.value or method == Method.DELETE.value:
                content_type = "application/json"
            else:
                content_type = "application/x-www-form-urlencoded"
            headers_: dict = {"Content-type": content_type}
        if not allow_redirection_:
            headers_["allow_redirection"] = allow_redirection_
        if isinstance(headers, dict):
            headers_.update(headers)
        all_params["headers"] = headers_

    @staticmethod
    def __build_optional_args(all_args: dict) -> dict:
        optional_args: dict = {}
        for key in optional_args_keys:
            optional_args[key] = all_args.get(key)
        if optional_args:
            optional_args_copy: dict = deepcopy(optional_args)
            for k, v in optional_args_copy.items():
                if not v and k in optional_args:
                    del optional_args[k]
        return optional_args

    @staticmethod
    def __build_api_name(api_name: str, func: callable) -> str:
        if isinstance(api_name, str) and api_name.strip() != "":
            return api_name
        return func.__name__


__all__ = ["Rest", "Method"]
