#!/usr/bin/env python
# -*- coding:utf-8 -*-
import os
import platform
import sys
import traceback
from datetime import datetime
from functools import wraps
from logging import StreamHandler, Formatter, Logger, Handler, Filter, LoggerAdapter
from logging.handlers import RotatingFileHandler
from pathlib import Path
from threading import Lock
from typing import TypeVar
from threading import current_thread

from .level import LogLevel
from ..converter.storage import StorageUnit
from ..config import LogConfig
from ..utils.computer import Disk
from ..utils.tools import generate_random_str

from .. import _local_var

K = TypeVar("K")
V = TypeVar("V")


class _LoggerWrapper(Logger):
    def __init__(self, name: str):
        super().__init__(name, LogConfig.level.value)
        self.__initialize()

    def __initialize(self):
        log_dir: Path = Path(LogConfig.dir)
        if not log_dir.exists():
            log_dir.mkdir(parents=True)
        if not LogConfig.off:
            for handler_type in ["ch", "fh"]:
                if handler_type == "fh" and LogConfig.off_file:
                    continue
                elif handler_type == "ch" and LogConfig.off_console:
                    continue
                self.addHandler(self.__handler__(handler_type))
        self.setLevel(LogConfig.level.value)

    @staticmethod
    def __handler__(handlerType: str) -> Handler:
        if "ch" == handlerType:
            handler = StreamHandler()
            level_filter = Filter()
            level: LogLevel = LogConfig.level_console if LogConfig.level_console else LogLevel.CRITICAL
            level_filter.filter = lambda record: record.levelno <= level.value
            handler.addFilter(level_filter)
        elif "fh" == handlerType:
            handler = RotatingFileHandler(filename=LogConfig.path,
                                          encoding=LogConfig.coding,
                                          maxBytes=LogConfig.max_bytes,
                                          backupCount=LogConfig.backup_count)
            level_filter = Filter()
            level: LogLevel = LogConfig.level_file if LogConfig.level_file else LogLevel.CRITICAL
            level_filter.filter = lambda record: record.levelno <= level.value
            handler.addFilter(level_filter)
        else:
            raise Exception(f"unknown handler type: {handlerType}")

        handler.setLevel(LogConfig.level.value)
        handler.setFormatter(Formatter(LogConfig.format))
        return handler


class _LoggerAdapterWrapper(LoggerAdapter):
    __lock = Lock()
    __instance = None

    def __new__(cls, *args, **kwargs):
        with cls.__lock:
            if not cls.__instance:
                cls.__instance = object.__new__(cls)
        return cls.__instance

    def process(self, msg, kwargs):
        key = current_thread().ident
        thread_local = _local_var.__dict__
        if key not in thread_local or thread_local[key] is None:
            msg = f": {msg}"
            return msg, kwargs
        msg = f"[{thread_local[key]}]: {msg}"
        return super().process(msg, kwargs)

    def trace(self, value: V = None):
        def __inner(func):
            @wraps(func)
            def __wrapper(*args, **kwargs):
                if not value:
                    _value = generate_random_str(16)
                else:
                    _value = value
                set_ok = self.__set(_value)
                try:
                    return func(*args, **kwargs)
                except BaseException as e:
                    msg = f"\nexception trace: {self.get_trace_id()}\n{'EXCEPTION STACK'.center(35, '*')}\n{str(e)}"
                    self.error(f"{msg}\n{traceback.format_exc()}")
                    e.__init__(msg)
                    raise
                finally:
                    if set_ok:
                        self.__remove()

            return __wrapper

        return __inner

    @staticmethod
    def __set(value: V) -> bool:
        key = current_thread().ident
        thread_local = _local_var.__dict__
        if key and key not in thread_local:
            thread_local[key] = value
            return True
        return False

    @staticmethod
    def get_trace_id(default: V = None) -> V:
        key = current_thread().ident
        thread_local = _local_var.__dict__
        if key in thread_local:
            return thread_local[key]
        return default

    @staticmethod
    def __remove():
        key = current_thread().ident
        thread_local = _local_var.__dict__
        if key in thread_local:
            del thread_local[key]


class _LoggerFactory(object):
    __lock = Lock()
    __instance = None

    def __new__(cls, *args, **kwargs):
        with cls.__lock:
            if not cls.__instance:
                cls.__instance = object.__new__(cls)
        return cls.__instance

    def __init__(self):
        self.__show_banner()

    def __show_banner(self):
        if not LogConfig.banner:
            return
        init_len = 140
        split_len = 32
        key_len = 16
        value_len = 0
        left_len = 4
        logo = 'INITIALIZE LOG ENVIRONMENT'
        logo_len = len(logo)
        content = f"""
{''.center(init_len, "#")}
{'#'.ljust(int((init_len - logo_len) / 2), " ")}{logo}{'#'.rjust(int((init_len - logo_len) / 2), " ")}
"""
        disk = Disk(str(LogConfig.dir))
        banner_dict = {
            "Python Version": platform.python_version(),
            "Python Compiler": platform.python_compiler(),
            "System Encoding": sys.getdefaultencoding(),
            "OS Version": platform.platform(),
            "CPU": platform.processor(),
            "Disk Free": f"{disk.get_free(StorageUnit.MB)} MB ({disk.get_free_percent()}%) ",
            "Date Time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "Weekend": f"{datetime.now().strftime('%A')}",
            "Log Path": LogConfig.path,
            "Log Level": LogConfig.level,
            "Log Back Content": LogConfig.backup_count,
            "Log Max Bytes": LogConfig.max_bytes,
            "Log Off file": f"{LogConfig.off_file} (Turn off log output to file)",
            "Log Off console": f"{LogConfig.off_console} (Turn off log output to console)",
            "Log Off": f"{LogConfig.off} (Turn off all log output)"
        }

        for k, v in banner_dict.items():
            v_ = str(v)
            line = f"{k.ljust(key_len)}{'=>'.center(split_len, ' ')}{v_.rjust(value_len)}"
            content += f"{'#'.ljust(left_len, ' ')}{line}{'#'.rjust(init_len - len(line) - left_len, ' ')}\n"

        content += f"{''.center(init_len, '#')}"
        logger = self.get_logger()
        logger.info(content)

    def get_logger(self) -> '_LoggerAdapterWrapper':
        logger = _LoggerWrapper("root")
        return _LoggerAdapterWrapper(logger, {})


LoggerFactory = _LoggerFactory()

__all__ = ["LoggerFactory"]
