#!/usr/bin/env python
# -*- coding:utf-8 -*-
import warnings
from typing import Generic, TypeVar, Callable, List, Optional, Any, Iterable

from ..collection import ArrayList
from ..exceptions import raise_exception, TypeException, NonePointerException

T = TypeVar('T', Iterable, Any)
U = TypeVar('U', Iterable, Any)
L = TypeVar('L')


class Optionals(Generic[T]):

    def __init__(self, value: T):
        self.__value = value

    @staticmethod
    def of(value: T) -> 'Optionals[T]':
        if not value:
            raise_exception(NonePointerException("value can't none"))
        return Optionals(value)

    @staticmethod
    def of_none_able(value) -> 'Optionals[T]':
        return Optionals.of(value) if value else Optionals(None)

    def get(self) -> Optional[T] or Optional[Iterable[T]]:
        if not self.__value:
            raise_exception(NonePointerException("No value present"))
        return self.__value

    def is_present(self) -> bool:
        return self.__value is not None

    def is_empty(self) -> bool:
        return self.__value is None

    def if_present(self, action: Callable[[T], None]) -> None:
        if self.__value:
            action(self.__value)

    def if_present_or_else(self, action: Callable[[T], None], empty_action: Callable) -> None:
        if self.__value:
            action(self.__value)
        else:
            empty_action(self.__value)

    def filter(self, predicate: Callable[[T], bool]) -> 'Optionals[T]':
        if not self.is_present():
            return self
        else:
            return self if predicate(self.__value) else Optionals(None)

    def map(self, mapper: Callable[[T], U]) -> 'Optionals[U]':
        if not self.is_present():
            return Optionals(None)
        else:
            return Optionals.of_none_able(mapper(self.__value))

    def flat_map(self, mapper: Callable[[T], U]) -> 'Optionals[U]':
        if not self.is_present():
            return Optionals(None)
        else:
            ret: Optionals[U] = mapper(self.__value)
            if not ret:
                raise_exception(NonePointerException("value can't none"))
            return ret

    def or_(self, supplier: Callable[[], 'Optionals[T]']) -> 'Optionals[T]':
        if self.is_present():
            return self
        else:
            ret = supplier()
            if not ret:
                raise_exception(NonePointerException("value can't none"))
            return ret

    def or_else(self, other: T) -> T:
        return self.__value if self.__value else other

    def or_else_get(self, supplier: Callable[[], 'Optionals[T]']) -> Optional[T]:
        return self.__value if self.__value else supplier()

    def or_else_raise(self, exception: BaseException = None) -> Optional[T]:
        if not self.__value:
            if exception:
                raise_exception(exception)
            else:
                raise_exception(NonePointerException("No value present"))
        return self.__value

    def stream(self) -> ArrayList[L]:
        warnings.warn('Stream is not recommended. The type information of the element is lost. '
                      'Until the type loss issue is fixed, it will remain in the DeprecationWarning state',
                      DeprecationWarning, stacklevel=2)
        if issubclass(type(self.__value), List):
            return ArrayList(iterable=self.__value)
        raise_exception(TypeException(f"expect iterable object, got {type(self.__value).__name__}"))

